@@ -7,12 +7,14 @@ setuptools entry points.
 
 from __future__ import absolute_import
 from __future__ import unicode_literals
+from __future__ import print_function
 
 # import udiskie.depend first - for side effects!
-from .depend import has_Notify, has_Gtk
+from .depend import has_Notify, has_Gtk, _in_X
 
+import sys
 import inspect
-import logging
+import logging.config
 import traceback
 
 from docopt import docopt, DocoptExit
@@ -22,8 +24,9 @@ from gi.repository import GLib
 import udiskie
 import udiskie.config
 import udiskie.mount
+import udiskie.compat
 from .async_ import AsyncList, Coroutine, Return, RunForever
-from .common import extend, str2unicode
+from .common import extend, str2unicode, ObjDictView
 from .locale import _
 
 
@@ -34,6 +37,15 @@ __all__ = [
 ]
 
 
+def deprecation_warning(text):
+    """Show a deprecation warning."""
+    # NOTE: not using `warnings.warn(text, DeprecationWarning)`, because that
+    # requires starting the python interpreter to be started with `-Wd`, which
+    # in turn shows more warnings about removal of Gtk.StatusIcon.
+    log = logging.getLogger("udiskie.DeprecationWarning")
+    log.warning(_("Deprecation warning: {}", text))
+
+
 @Coroutine.from_generator_function
 def get_backend(version=None):
     """
@@ -44,8 +56,8 @@ def get_backend(version=None):
     :raises GLib.GError: if unable to connect to UDisks dbus service.
     :raises ValueError: if the version is invalid
 
-    If ``version`` has a false truth value, try to connect to UDisks1 and
-    fall back to UDisks2 if not available.
+    If ``version`` has a false truth value, try to connect to UDisks2 and
+    fall back to UDisks1 if not available.
     """
     if not version:
         try:
@@ -58,6 +70,9 @@ def get_backend(version=None):
     elif version == 1:
         import udiskie.udisks1
         daemon = yield udiskie.udisks1.Daemon.create()
+        deprecation_warning(_(
+            'Using UDisks1. Support will be discontinued '
+            'in the next major version of udiskie.'))
     elif version == 2:
         import udiskie.udisks2
         daemon = yield udiskie.udisks2.Daemon.create()
@@ -119,6 +134,13 @@ class OptionalValue(object):
         return self._choice(args) and str2unicode(args[self._name])
 
 
+class SelectLevel(logging.Filter):
+    def __init__(self, level):
+        self.level = level
+    def filter(self, record):
+        return record.levelno == self.level
+
+
 class _EntryPoint(object):
 
     """
@@ -158,17 +180,43 @@ class _EntryPoint(object):
 
         :param list argv: command line parameters
         """
+        udiskie.compat.patch_print_unicode()
         # parse program options (retrieve log level and config file name):
         args = docopt(self.usage, version=self.name + ' ' + self.version)
         default_opts = self.option_defaults
         program_opts = self.program_options(args)
         # initialize logging configuration:
         log_level = program_opts.get('log_level', default_opts['log_level'])
-        if log_level <= logging.DEBUG:
-            fmt = _('%(levelname)s [%(asctime)s] %(name)s: %(message)s')
-        else:
-            fmt = _('%(message)s')
-        logging.basicConfig(level=log_level, format=fmt)
+        debug = log_level <= logging.DEBUG
+        logging.config.dictConfig({
+            'version': 1,
+            'disable_existing_loggers': False,
+            'formatters': {
+                'plain':  {'format': _('%(message)s')},
+                'detail': {'format': _('%(levelname)s [%(asctime)s] %(name)s: %(message)s')},
+            },
+            'filters': {
+                'info': {'()': 'udiskie.cli.SelectLevel', 'level': logging.INFO},
+            },
+            'handlers': {
+                'info':  {'class': 'logging.StreamHandler',
+                          'stream': 'ext://sys.stdout',
+                          'formatter': 'plain',
+                          'filters': ['info']},
+                'error': {'class': 'logging.StreamHandler',
+                          'stream': 'ext://sys.stderr',
+                          'formatter': 'plain',
+                          'level': 'WARNING'},
+                'debug': {'class': 'logging.StreamHandler',
+                          'stream': 'ext://sys.stderr',
+                          'formatter': 'detail'},
+            },
+            # configure root logger:
+            'root': {
+                'handlers': ['info', 'debug' if debug else 'error'],
+                'level': log_level,
+            },
+        })
         # parse config options
         config_file = OptionalValue('--config')(args)
         config = udiskie.config.Config.from_file(config_file)
@@ -180,6 +228,11 @@ class _EntryPoint(object):
         self.config = config
         self.options = options
         self.exit_status = 0
+        if sys.version_info < (3,5):
+            deprecation_warning(_(
+                "Running on python {}.{}. The next major version of udiskie "
+                "will require at least python 3.5!",
+                *sys.version_info[:2]))
 
     def program_options(self, args):
         """
@@ -251,7 +304,9 @@ class _EntryPoint(object):
         """Start asynchronous operations."""
         try:
             self.udisks = yield get_backend(self.options['udisks_version'])
-            yield self._init()
+            results = yield self._init()
+            if not all(results):
+                self.exit_status = 1
         except Exception:
             self.exit_status = 1
             # Print the stack trace only up to the current level:
@@ -259,6 +314,33 @@ class _EntryPoint(object):
         self.mainloop.quit()
 
 
+class Component(object):
+
+    def __init__(self, create):
+        self.create = create
+        self.instance = None
+
+    @property
+    def active(self):
+        return self.instance is not None and self.instance.active
+
+    def activate(self):
+        if self.instance is None:
+            self.instance = self.create()
+        if not self.instance.active:
+            self.instance.activate()
+
+    def deactivate(self):
+        if self.active:
+            self.instance.deactivate()
+
+    def toggle(self):
+        if self.active:
+            self.deactivate()
+        else:
+            self.activate()
+
+
 class Daemon(_EntryPoint):
 
     """
@@ -292,6 +374,10 @@ class Daemon(_EntryPoint):
         -t, --tray                              Show tray icon
         -s, --smart-tray                        Auto hide tray icon
         -T, --no-tray                           Disable tray icon
+        -m MENU, --menu MENU                    Tray menu [flat/nested]
+
+        --appindicator                          Use appindicator for status icon
+        --no-appindicator                       Don't use appindicator
 
         --password-cache MINUTES                Set password cache timeout
         --no-password-cache                     Disable password cache
@@ -299,6 +385,9 @@ class Daemon(_EntryPoint):
         -p COMMAND, --password-prompt COMMAND   Command for password retrieval
         -P, --no-password-prompt                Disable unlocking
 
+        --notify-command COMMAND                Command to execute on events
+        --no-notify-command                     Disable command notifications
+
     Deprecated options:
         -f PROGRAM, --file-manager PROGRAM      Set program for browsing
         -F, --no-file-manager                   Disable browsing
@@ -310,9 +399,12 @@ class Daemon(_EntryPoint):
         'automount': True,
         'notify': True,
         'tray': False,
+        'menu': 'flat',
+        'appindicator': False,
         'file_manager': 'xdg-open',
         'password_prompt': 'builtin:gui',
         'password_cache': False,
+        'notify_command': None,
     })
 
     option_rules = extend(_EntryPoint.option_rules, {
@@ -322,9 +414,12 @@ class Daemon(_EntryPoint):
             '--tray': True,
             '--no-tray': False,
             '--smart-tray': 'auto'}),
+        'menu': Value('--menu'),
+        'appindicator': Switch('appindicator'),
         'file_manager': OptionalValue('--file-manager'),
         'password_prompt': OptionalValue('--password-prompt'),
         'password_cache': OptionalValue('--password-cache'),
+        'notify_command': OptionalValue('--notify-command'),
     })
 
     def _init(self):
@@ -336,6 +431,7 @@ class Daemon(_EntryPoint):
         config = self.config
         options = self.options
 
+        # prepare mounter object
         prompt = udiskie.prompt.password(options['password_prompt'])
         browser = udiskie.prompt.browser(options['file_manager'])
         cache = None
@@ -345,14 +441,14 @@ class Daemon(_EntryPoint):
             timeout = int(options['password_cache']) * 60
             cache = udiskie.cache.PasswordCache(timeout)
 
-        mounter = udiskie.mount.Mounter(
-            mount_options=config.mount_options,
-            ignore_device=config.ignore_device,
+        self.mounter = udiskie.mount.Mounter(
+            config=config.device_config,
             prompt=prompt,
             browser=browser,
             cache=cache,
             udisks=self.udisks)
 
+        # check component availability
         if options['notify'] and not has_Notify():
             libnotify_not_available = _(
                 "Typelib for 'libnotify' is not available. Possible causes include:"
@@ -363,66 +459,100 @@ class Daemon(_EntryPoint):
             logging.getLogger(__name__).error(libnotify_not_available)
             options['notify'] = False
 
+        if options['tray'] and not has_Gtk(3) and not _in_X:
+            no_X_session = _(
+                "Not run within X session. "
+                "\nStarting udiskie without tray icon.\n")
+            logging.getLogger(__name__).error(no_X_session)
+            options['tray'] = False
+
         if options['tray'] and not has_Gtk(3):
             gtk3_not_available = _(
                 "Typelib for 'Gtk 3.0' is not available. Possible causes include:"
                 "\n\t- GTK3 is not installed"
                 "\n\t- the typelib is provided by a separate package"
                 "\n\t- GTK3 was built with introspection disabled"
-                "\n\nStarting udiskie without tray icon.")
+                "\nStarting udiskie without tray icon.\n")
             logging.getLogger(__name__).error(gtk3_not_available)
             options['tray'] = False
 
+        # start components
         tasks = []
 
-        # notifications (optional):
+        self.notify         = Component(self._load_notify)
+        self.statusicon     = Component(self._load_statusicon)
+        self.automounter    = Component(self._load_automounter)
+
         if options['notify']:
-            import udiskie.notify
-            from gi.repository import Notify
-            Notify.init('udiskie')
-            aconfig = config.notification_actions
-            if options['automount']:
-                aconfig.setdefault('device_added', [])
-            else:
-                aconfig.setdefault('device_added', ['mount'])
-            udiskie.notify.Notify(Notify.Notification.new,
-                                  mounter=mounter,
-                                  timeout=config.notifications,
-                                  aconfig=aconfig)
-
-        # tray icon (optional):
+            self.notify.activate()
+        if options['notify_command']:
+            # is currently enabled/disabled statically only once:
+            self.notify_command()
         if options['tray']:
-            import udiskie.tray
-            tray_classes = {True: udiskie.tray.TrayIcon,
-                            'auto': udiskie.tray.AutoTray}
-            if options['tray'] not in tray_classes:
-                raise ValueError("Invalid tray: %s" % (options['tray'],))
-            icons = udiskie.tray.Icons(config.icon_names)
-            actions = udiskie.mount.DeviceActions(mounter)
-            menu_maker = udiskie.tray.SmartUdiskieMenu(
-                mounter,
-                icons,
-                actions,
-                quit_action=self.mainloop.quit)
-            TrayIcon = tray_classes[options['tray']]
-            statusicon = TrayIcon(menu_maker, icons)
-            tasks.append(statusicon.task)
+            self.statusicon.activate()
+            tasks.append(self.statusicon.instance._icon.task)
         else:
-            statusicon = None
             tasks.append(RunForever)
-
-        # automounter
         if options['automount']:
-            import udiskie.automount
-            udiskie.automount.AutoMounter(mounter)
-            tasks.append(mounter.add_all())
-
-        # Note: mounter and statusicon are saved so these are kept alive:
-        self.mounter = mounter
-        self.statusicon = statusicon
+            self.automounter.activate()
+            tasks.append(self.mounter.add_all())
 
         return AsyncList(tasks)
 
+    def _load_notify(self):
+        import udiskie.notify
+        from gi.repository import Notify
+        Notify.init('udiskie')
+        aconfig = self.config.notification_actions
+        if self.options['automount']:
+            aconfig.setdefault('device_added', [])
+        else:
+            aconfig.setdefault('device_added', ['mount'])
+        return udiskie.notify.Notify(
+            Notify.Notification.new,
+            mounter=self.mounter,
+            timeout=self.config.notifications,
+            aconfig=aconfig)
+
+    def notify_command(self):
+        import udiskie.prompt
+        return udiskie.prompt.notify_command(
+            self.options['notify_command'], self.mounter)
+
+    def _load_statusicon(self):
+        import udiskie.tray
+        options = self.options
+
+        if options['tray'] == 'auto':
+            smart = True
+        elif options['tray'] is True:
+            smart = False
+        else:
+            raise ValueError("Invalid tray: %s" % (options['tray'],))
+        icons = udiskie.tray.Icons(self.config.icon_names)
+        actions = udiskie.mount.DeviceActions(self.mounter)
+
+        if options['menu'] == 'flat':
+            flat = True
+        # dropped legacy 'nested' mode:
+        elif options['menu'] in ('smart', 'nested'):
+            flat = False
+        else:
+            raise ValueError("Invalid menu: %s" % (options['menu'],))
+
+        menu_maker = udiskie.tray.UdiskieMenu(self, icons, actions, flat)
+        if options['appindicator']:
+            import udiskie.appindicator
+            TrayIcon = udiskie.appindicator.AppIndicatorIcon
+        else:
+            TrayIcon = udiskie.tray.TrayIcon
+        trayicon = TrayIcon(menu_maker, icons)
+        return udiskie.tray.UdiskieStatusIcon(trayicon, menu_maker, smart)
+
+    def _load_automounter(self):
+        import udiskie.automount
+        return udiskie.automount.AutoMounter(self.mounter)
+
 
 class Mount(_EntryPoint):
 
@@ -462,7 +592,7 @@ class Mount(_EntryPoint):
     name = 'udiskie-mount'
 
     option_defaults = extend(_EntryPoint.option_defaults, {
-        'recursive': False,
+        'recursive': None,
         'options': None,
         '<device>': None,
         'password_prompt': 'builtin:tty',
@@ -484,16 +614,15 @@ class Mount(_EntryPoint):
         config = self.config
         options = self.options
 
+        device_config = config.device_config
         if options['options']:
-            opts = [o.strip() for o in options['options'].split(',')]
-            mount_options = lambda dev: opts
-        else:
-            mount_options = config.mount_options
+            device_config._filters.insert(0, udiskie.config.MountOptions({
+                'options': [o.strip() for o in options['options'].split(',')],
+            }))
 
         prompt = udiskie.prompt.password(options['password_prompt'])
         mounter = udiskie.mount.Mounter(
-            mount_options=mount_options,
-            ignore_device=config.ignore_device,
+            config=config.device_config,
             prompt=prompt,
             udisks=self.udisks)
 
@@ -501,10 +630,9 @@ class Mount(_EntryPoint):
         if options['<device>']:
             tasks = [mounter.add(path, recursive=recursive)
                      for path in options['<device>']]
-            # TODO: AND results
-            return AsyncList(tasks)
         else:
-            return mounter.add_all(recursive=recursive)
+            tasks = [mounter.add_all(recursive=recursive)]
+        return AsyncList(tasks)
 
 
 class Umount(_EntryPoint):
@@ -549,7 +677,7 @@ class Umount(_EntryPoint):
     name = 'udiskie-umount'
 
     option_defaults = extend(_EntryPoint.option_defaults, {
-        'detach': False,
+        'detach': None,
         'eject': False,
         'force': False,
         'lock': True,
@@ -568,8 +696,12 @@ class Umount(_EntryPoint):
 
         """Implements _EntryPoint._init."""
 
+        config = self.config
         options = self.options
-        mounter = udiskie.mount.Mounter(self.udisks)
+
+        mounter = udiskie.mount.Mounter(
+            self.udisks,
+            config=config.device_config)
 
         strategy = dict(detach=options['detach'],
                         eject=options['eject'],
@@ -578,7 +710,108 @@ class Umount(_EntryPoint):
             strategy['force'] = options['force']
             tasks = [mounter.remove(path, **strategy)
                      for path in options['<device>']]
-            # TODO: AND results
-            return AsyncList(tasks)
         else:
-            return mounter.remove_all(**strategy)
+            tasks = [mounter.remove_all(**strategy)]
+        return AsyncList(tasks)
+
+
+def _parse_filter(spec):
+    try:
+        key, val = spec.split('=', 1)
+    except ValueError:
+        if spec.startswith('!'):
+            val = False
+            key = spec[1:]
+        else:
+            val = True
+            key = spec
+    return key, val
+
+
+class Info(_EntryPoint):
+
+    """
+    udiskie-info: get information about handleable devices.
+
+    Usage:
+        udiskie-info [options] [-o OUTPUT] [-f FILTER]... (-a | DEVICE...)
+        udiskie-info (--help | --version)
+
+    General options:
+        -c FILE, --config=FILE      Set config file
+        -C, --no-config             Don't use config file
+
+        -v, --verbose               Increase verbosity (DEBUG)
+        -q, --quiet                 Decrease verbosity
+
+        -0, --udisks-auto           Auto discover UDisks version
+        -1, --use-udisks1           Use UDisks1 as backend
+        -2, --use-udisks2           Use UDisks2 as backend
+
+        -h, --help                  Show this help
+        -V, --version               Show version information
+
+    Unmount options:
+        -a, --all                   List all handleable devices
+
+        -o COL, --output COL        Specify output columns in a format string
+                                    containing the allowed device attributes,
+                                    e.g.: "{ui_label} {is_luks}"
+                                    [default: device_presentation].
+
+        -f FILT, --filter FILT      Print only devices that match the given
+                                    filter.
+    """
+
+    name = 'udiskie-info'
+
+    option_defaults = extend(_EntryPoint.option_defaults, {
+        'output': '',
+        'filter': '',
+        '<device>': None,
+    })
+
+    option_rules = extend(_EntryPoint.option_rules, {
+        'output': Value('--output'),
+        'filter': Value('--filter'),
+        '<device>': Value('DEVICE'),
+    })
+
+    def _init(self):
+
+        """Implements _EntryPoint._init."""
+
+        config = self.config
+        options = self.options
+
+        mounter = udiskie.mount.Mounter(
+            self.udisks,
+            config=config.device_config)
+
+        if options['<device>']:
+            devices = [self.udisks.find(path) for path in options['<device>']]
+        else:
+            devices = mounter.get_all_handleable()
+
+        DeviceFilter = udiskie.config.DeviceFilter
+        output = options['output']
+        # old behaviour: single attribute
+        if output in DeviceFilter.VALID_PARAMETERS:
+            def format_output(device):
+                return getattr(device, output)
+        # new behaviour: format string
+        else:
+            from string import Formatter
+            formatter = Formatter()
+            def format_output(device):
+                view = ObjDictView(device, DeviceFilter.VALID_PARAMETERS)
+                return formatter.vformat(output, (), view)
+
+        filters = [_parse_filter(spec) for spec in options['filter']]
+        matcher = DeviceFilter(dict(filters))
+
+        for device in devices:
+            if matcher.match(device):
+                print(format_output(device))
+
+        return AsyncList([])
