@@ -5,13 +5,16 @@ Mount utilities.
 from __future__ import absolute_import
 from __future__ import unicode_literals
 
+from distutils.spawn import find_executable
 from collections import namedtuple
 from functools import partial
+import inspect
 import logging
+import os
 
-from .async_ import AsyncList, Coroutine, Return
+from .async_ import AsyncList, Coroutine, Return, sleep
 from .common import wraps, setdefault, exc_message
-from .config import IgnoreDevice, FilterMatcher
+from .config import IgnoreDevice, match_config
 from .locale import _
 
 
@@ -27,7 +30,7 @@ def _False():
     yield Return(False)
 
 
-def _find_device(fn, set_error=False):
+def _find_device(fn):
     """
     Decorator for Mounter methods taking a Device as their first argument.
 
@@ -45,6 +48,27 @@ def _find_device(fn, set_error=False):
     return wrapper
 
 
+def _find_device_auto_losetup(fn):
+    @Coroutine.from_generator_function
+    def wrapper(self, device_or_path, *args, **kwargs):
+        try:
+            try:
+                device = self.udisks.find(device_or_path)
+            except ValueError as e:
+                if not os.path.isfile(device_or_path):
+                    raise
+                device = yield self.losetup(device_or_path)
+                bound = inspect.getcallargs(fn, self, device, *args, **kwargs)
+                if bound.get('recursive') is False:
+                    yield Return(device)
+        except Exception as e:
+            self._log.error(exc_message(e))
+            yield Return(False)
+        result = yield Coroutine(fn(self, device, *args, **kwargs))
+        yield Return(result)
+    return wrapper
+
+
 def _sets_async_error(fn):
     @wraps(fn)
     def wrapper(self, device, *args, **kwargs):
@@ -76,6 +100,11 @@ def _is_parent_of(parent, child):
     return False
 
 
+def _get_parent(device):
+    """Return the container device or ``None``."""
+    return device.partition_slave or device.luks_cleartext_slave
+
+
 class Mounter(object):
 
     """
@@ -91,8 +120,7 @@ class Mounter(object):
     """
 
     def __init__(self, udisks,
-                 mount_options=None,
-                 ignore_device=None,
+                 config=None,
                  prompt=None,
                  browser=None,
                  cache=None):
@@ -100,7 +128,7 @@ class Mounter(object):
         Initialize mounter with the given defaults.
 
         :param udisks: udisks service object. May be a Sniffer or a Daemon.
-        :param FilterMatcher filter: customize mount options and handleability
+        :param list config: list of :class:`DeviceFilter`
         :param callable prompt: retrieve passwords for devices
         :param callable browser: open devices
 
@@ -108,9 +136,10 @@ class Mounter(object):
         If browser is None, browse will not work.
         """
         self.udisks = udisks
-        self._mount_options = mount_options or (lambda device: None)
-        self._ignore_device = ignore_device or FilterMatcher([], False)
-        self._ignore_device._filters += [
+        self._config = (config or []) + [
+            IgnoreDevice({'symlinks': '/dev/mapper/docker-*', 'ignore': True}),
+            IgnoreDevice({'symlinks': '/dev/disk/by-id/dm-name-docker-*', 'ignore': True}),
+            IgnoreDevice({'is_loop': True, 'loop_file': '/*', 'ignore': False}),
             IgnoreDevice({'is_block': False, 'ignore': True}),
             IgnoreDevice({'is_external': False, 'ignore': True}),
             IgnoreDevice({'is_ignored': True, 'ignore': True})]
@@ -170,14 +199,21 @@ class Mounter(object):
         if device.is_mounted:
             self._log.info(_('not mounting {0}: already mounted', device))
             yield Return(True)
-        fstype = str(device.id_type)
-        options = self._mount_options(device)
-        kwargs = dict(fstype=fstype, options=options)
+        options = match_config(self._config, device, 'options', None)
+        kwargs = dict(options=options)
         self._log.debug(_('mounting {0} with {1}', device, kwargs))
+        self._check_device_before_mount(device)
         mount_path = yield device.mount(**kwargs)
         self._log.info(_('mounted {0} on {1}', device, mount_path))
         yield Return(True)
 
+    def _check_device_before_mount(self, device):
+        if device.id_type == 'ntfs' and not find_executable('ntfs-3g'):
+            self._log.warn(_(
+                "Mounting NTFS device with default driver.\n"
+                "Please install 'ntfs-3g' if you experience problems or the "
+                "device is readonly."))
+
     @_sets_async_error
     @_find_device
     def unmount(self, device):
@@ -222,12 +258,19 @@ class Mounter(object):
         unlocked = yield self._unlock_from_cache(device)
         if unlocked:
             yield Return(True)
-        password = yield self._prompt(device)
+        unlocked = yield self._unlock_from_keyfile(device)
+        if unlocked:
+            yield Return(True)
+        password = yield self._prompt(device, self.udisks.keyfile_support)
         if password is None:
             self._log.debug(_('not unlocking {0}: cancelled by user', device))
             yield Return(False)
-        self._log.debug(_('unlocking {0}', device))
-        yield device.unlock(password)
+        if isinstance(password, bytes):
+            self._log.debug(_('unlocking {0} using keyfile', device))
+            yield device.unlock_keyfile(password)
+        else:
+            self._log.debug(_('unlocking {0}', device))
+            yield device.unlock(password)
         self._update_cache(device, password)
         self._log.info(_('unlocked {0}', device))
         yield Return(True)
@@ -246,7 +289,30 @@ class Mounter(object):
         except Exception:
             self._log.debug(_('failed to unlock {0} using cached password', device))
             yield Return(False)
-        self._log.debug(_('unlocked {0} using cached password', device))
+        self._log.info(_('unlocked {0} using cached password', device))
+        yield Return(True)
+
+    @Coroutine.from_generator_function
+    def _unlock_from_keyfile(self, device):
+        if not self.udisks.keyfile_support:
+            yield Return(False)
+        filename = match_config(self._config, device, 'keyfile', None)
+        if filename is None:
+            self._log.debug(_('No matching keyfile rule for {}.', device))
+            yield Return(False)
+        try:
+            with open(filename, 'rb') as f:
+                keyfile = f.read()
+        except IOError:
+            self._log.warn(_('configured keyfile for {0} not found', device))
+            yield Return(False)
+        self._log.debug(_('unlocking {0} using keyfile {1}', device, filename))
+        try:
+            yield device.unlock_keyfile(keyfile)
+        except Exception:
+            self._log.debug(_('failed to unlock {0} using keyfile', device))
+            yield Return(False)
+        self._log.info(_('unlocked {0} using keyfile', device))
         yield Return(True)
 
     def _update_cache(self, device, password):
@@ -283,8 +349,8 @@ class Mounter(object):
 
     # add/remove (unlock/lock or mount/unmount)
     @_suppress_error
-    @_find_device
-    def add(self, device, recursive=False):
+    @_find_device_auto_losetup
+    def add(self, device, recursive=None):
         """
         Mount or unlock the device depending on its type.
 
@@ -298,26 +364,29 @@ class Mounter(object):
         elif device.is_crypto:
             success = yield self.unlock(device)
             if success and recursive:
-                self.udisks._sync()
+                yield self.udisks._sync()
                 device = self.udisks[device.object_path]
                 success = yield self.add(
                     device.luks_cleartext_holder,
                     recursive=True)
-        elif recursive and device.is_partition_table:
-            tasks = []
-            for dev in self.get_all_handleable():
-                if dev.is_partition and dev.partition_slave == device:
-                    tasks.append(self.add(dev, recursive=True))
-            # TODO: AND results
-            success = yield AsyncList(tasks)
+        elif (recursive
+              and device.is_partition_table
+              and self.is_handleable(device)):
+            tasks = [
+                self.add(dev, recursive=True)
+                for dev in self.get_all_handleable()
+                if dev.is_partition and dev.partition_slave == device
+            ]
+            results = yield AsyncList(tasks)
+            success = all(results)
         else:
             self._log.info(_('not adding {0}: unhandled device', device))
             yield Return(False)
         yield Return(success)
 
     @_suppress_error
-    @_find_device
-    def auto_add(self, device, recursive=False):
+    @_find_device_auto_losetup
+    def auto_add(self, device, recursive=None):
         """
         Automatically attempt to mount or unlock a device, but be quiet if the
         device is not supported.
@@ -328,7 +397,9 @@ class Mounter(object):
         :rtype: bool
         """
         success = True
-        if not self.is_handleable(device):
+        if device.is_luks_cleartext and self.udisks.version_info >= (2,7,0):
+            yield sleep(1.5)    # temporary workaround for #153, unreliable...
+        if not self.is_automount(device):
             pass
         elif device.is_filesystem:
             if not device.is_mounted:
@@ -337,18 +408,19 @@ class Mounter(object):
             if self._prompt and not device.is_unlocked:
                 success = yield self.unlock(device)
             if success and recursive:
-                self.udisks._sync()
+                yield self.udisks._sync()
                 device = self.udisks[device.object_path]
                 success = yield self.auto_add(
                     device.luks_cleartext_holder,
                     recursive=True)
         elif recursive and device.is_partition_table:
-            tasks = []
-            for dev in self.get_all_handleable():
-                if dev.is_partition and dev.partition_slave == device:
-                    tasks.append(self.auto_add(dev, recursive=True))
-            # TODO: AND results
-            success = yield AsyncList(tasks)
+            tasks = [
+                self.auto_add(dev, recursive=True)
+                for dev in self.get_all_handleable()
+                if dev.is_partition and dev.partition_slave == device
+            ]
+            results = yield AsyncList(tasks)
+            success = all(results)
         else:
             self._log.debug(_('not adding {0}: unhandled device', device))
         yield Return(success)
@@ -369,23 +441,23 @@ class Mounter(object):
         :rtype: bool
         """
         if device.is_filesystem:
-            success = yield self.unmount(device)
+            if device.is_mounted or not device.is_loop or detach is False:
+                success = yield self.unmount(device)
         elif device.is_crypto:
             if force and device.is_unlocked:
                 yield self.auto_remove(device.luks_cleartext_holder, force=True)
             success = yield self.lock(device)
-        elif force and (device.is_partition_table or device.is_drive):
-            tasks = []
-            for child in self.get_all_handleable():
-                if _is_parent_of(device, child):
-                    tasks.append(self.auto_remove(
-                        child,
-                        force=True,
-                        detach=detach,
-                        eject=eject,
-                        lock=lock))
-            # TODO: AND results
-            success = yield AsyncList(tasks)
+        elif (force
+              and (device.is_partition_table or device.is_drive)
+              and self.is_handleable(device)):
+            kw = dict(force=True, detach=detach, eject=eject, lock=lock)
+            tasks = [
+                self.auto_remove(child, **kw)
+                for child in self.get_all_handleable()
+                if _is_parent_of(device, child)
+            ]
+            results = yield AsyncList(tasks)
+            success = all(results)
         else:
             self._log.info(_('not removing {0}: unhandled device', device))
             success = False
@@ -395,7 +467,9 @@ class Mounter(object):
             success = yield self.lock(device)
         if eject:
             success = yield self.eject(device)
-        if detach:
+        if (detach or detach is None) and device.is_loop:
+            success = yield self.delete(device, remove=False)
+        elif detach:
             success = yield self.detach(device)
         yield Return(success)
 
@@ -426,17 +500,14 @@ class Mounter(object):
             if device.is_unlocked:
                 success = yield self.lock(device)
         elif force and (device.is_partition_table or device.is_drive):
-            tasks = []
-            for child in self.get_all_handleable():
-                if _is_parent_of(device, child):
-                    tasks.append(self.auto_remove(
-                        child,
-                        force=True,
-                        detach=detach,
-                        eject=eject,
-                        lock=lock))
-            # TODO: AND results
-            success = yield AsyncList(tasks)
+            kw = dict(force=True, detach=detach, eject=eject, lock=lock)
+            tasks = [
+                self.auto_remove(child, **kw)
+                for child in self.get_all_handleable()
+                if _is_parent_of(device, child)
+            ]
+            results = yield AsyncList(tasks)
+            success = all(results)
         else:
             self._log.debug(_('not removing {0}: unhandled device', device))
         # if these operations work, everything is fine, we can return True:
@@ -445,7 +516,9 @@ class Mounter(object):
             success = yield self.lock(device)
         if eject and device.has_media:
             success = yield self.eject(device)
-        if detach and device.is_detachable:
+        if (detach or detach is None) and device.is_loop:
+            success = yield self.delete(device, remove=False)
+        elif detach and device.is_detachable:
             success = yield self.detach(device)
         yield Return(success)
 
@@ -469,7 +542,9 @@ class Mounter(object):
             self._log.warn(_('not ejecting {0}: drive not ejectable', drive))
             yield Return(False)
         if force:
-            yield self.auto_remove(drive, force=True)
+            # Can't autoremove 'device.drive', because that will be filtered
+            # due to block=False:
+            yield self.auto_remove(device.root, force=True)
         self._log.debug(_('ejecting {0}', device))
         yield drive.eject()
         self._log.info(_('ejected {0}', device))
@@ -501,6 +576,7 @@ class Mounter(object):
         yield Return(True)
 
     # mount_all/unmount_all
+    @Coroutine.from_generator_function
     def add_all(self, recursive=False):
         """
         Add all handleable devices that available at start.
@@ -509,12 +585,13 @@ class Mounter(object):
         :returns: whether all attempted operations succeeded
         :rtype: bool
         """
-        tasks = []
-        for device in self.udisks:
-            tasks.append(self.auto_add(device, recursive=recursive))
-        # TODO: AND results
-        return AsyncList(tasks)
+        tasks = [self.auto_add(device, recursive=recursive)
+                 for device in self.get_all_handleable_leaves()]
+        results = yield AsyncList(tasks)
+        success = all(results)
+        yield Return(success)
 
+    @Coroutine.from_generator_function
     def remove_all(self, detach=False, eject=False, lock=False):
         """
         Remove all filesystems handleable by udiskie.
@@ -525,14 +602,72 @@ class Mounter(object):
         :returns: whether all attempted operations succeeded
         :rtype: bool
         """
-        tasks = []
-        remove_args = dict(force=True, detach=detach, eject=eject, lock=lock)
-        for device in self.get_all_handleable():
-            if device.parent_object_path != '/':
-                continue
-            tasks.append(self.auto_remove(device, **remove_args))
-        # TODO: AND results
-        return AsyncList(tasks)
+        kw = dict(force=True, detach=detach, eject=eject, lock=lock)
+        tasks = [self.auto_remove(device, **kw)
+                 for device in self.get_all_handleable_roots()]
+        results = yield AsyncList(tasks)
+        success = all(results)
+        yield Return(success)
+
+    # loop devices
+    @Coroutine.from_generator_function
+    def losetup(self, image,
+                read_only=True, offset=None, size=None, no_part_scan=None):
+        """
+        Setup a loop device.
+
+        :param str image: path of the image file
+        :param bool read_only:
+        :param int offset:
+        :param int size:
+        :param bool no_part_scan:
+        :returns: the device object for the loop device
+        """
+        try:
+            device = self.udisks.find(image)
+        except ValueError:
+            pass
+        else:
+            self._log.info(_('not setting up {0}: already up', device))
+            yield Return(device)
+        if not os.path.isfile(image):
+            self._log.error(_('not setting up {0}: not a file', image))
+            yield Return(None)
+        if not self.udisks.loop_support:
+            self._log.error(_('not setting up {0}: unsupported in UDisks1', image))
+            yield Return(None)
+        self._log.debug(_('setting up {0}', image))
+        fd = os.open(image, os.O_RDONLY)
+        device = yield self.udisks.loop_setup(fd, {
+            'offset': offset,
+            'size': size,
+            'read-only': read_only,
+            'no-part-scan': no_part_scan,
+        })
+        self._log.info(_('set up {0} as {1}', image,
+                         device.device_presentation))
+        yield Return(device)
+
+    @_sets_async_error
+    @_find_device
+    def delete(self, device, remove=True):
+        """
+        Detach the loop device.
+
+        :param device: device object, block device path or mount path
+        :param bool remove: whether to unmount the partition etc.
+        :returns: whether the loop device is deleted
+        :rtype: bool
+        """
+        if not self.is_handleable(device) or not device.is_loop:
+            self._log.warn(_('not deleting {0}: unhandled device', device))
+            yield Return(False)
+        if remove:
+            yield self.auto_remove(device, force=True)
+        self._log.debug(_('deleting {0}', device))
+        yield device.delete()
+        self._log.info(_('deleted {0}', device))
+        yield Return(True)
 
     # iterate devices
     def is_handleable(self, device):
@@ -547,13 +682,25 @@ class Mounter(object):
         Currently this just means that the device is removable and holds a
         filesystem or the device is a LUKS encrypted volume.
         """
-        return not self._ignore_device(device)
+        ignored = self._ignore_device(device)
+        # propagate handleability of parent devices:
+        if ignored is None and device is not None:
+            return self.is_handleable(_get_parent(device))
+        return not ignored
+
+    def is_automount(self, device):
+        if not self.is_handleable(device):
+            return False
+        return match_config(self._config, device, 'automount', True)
+
+    def _ignore_device(self, device):
+        return match_config(self._config, device, 'ignore', False)
 
     def is_addable(self, device):
         """
         Check if device can be added with ``auto_add``.
         """
-        if not self.is_handleable(device):
+        if not self.is_automount(device):
             return False
         if device.is_filesystem:
             return not device.is_mounted
@@ -586,18 +733,90 @@ class Mounter(object):
         Enumerate all handleable devices currently known to udisks.
 
         :returns: handleable devices
-        :rtype: iterable
-
-        NOTE: returns only devices that are still valid. This protects from
-        race conditions inside udiskie.
+        :rtype: list
         """
-        return filter(self.is_handleable, self.udisks)
+        nodes = self.get_device_tree()
+        return [node.device
+                for node in sorted(nodes.values(), key=DevNode._sort_key)
+                if not node.ignored and node.device]
+
+    def get_all_handleable_roots(self):
+        """
+        Get list of all handleable devices, return only those that represent
+        root nodes within the filtered device tree.
+        """
+        nodes = self.get_device_tree()
+        return [node.device
+                for node in sorted(nodes.values(), key=DevNode._sort_key)
+                if not node.ignored and node.device
+                and (node.root == '/' or nodes[node.root].ignored)]
+
+    def get_all_handleable_leaves(self):
+        """
+        Get list of all handleable devices, return only those that represent
+        leaf nodes within the filtered device tree.
+        """
+        nodes = self.get_device_tree()
+        return [node.device
+                for node in sorted(nodes.values(), key=DevNode._sort_key)
+                if not node.ignored and node.device
+                and all(child.ignored for child in node.children)]
+
+    def get_device_tree(self):
+        """Get a tree of all devices."""
+        root = DevNode(None, None, [], None)
+        device_nodes = {
+            dev.object_path: DevNode(dev, dev.parent_object_path, [],
+                                     self._ignore_device(dev))
+            for dev in self.udisks
+        }
+        for node in device_nodes.values():
+            device_nodes.get(node.root, root).children.append(node)
+        device_nodes['/'] = root
+        for node in device_nodes.values():
+            node.children.sort(key=DevNode._sort_key)
+        # use parent as fallback, update top->down:
+        def propagate_ignored(node):
+            for child in node.children:
+                if child.ignored is None:
+                    child.ignored = node.ignored
+                propagate_ignored(child)
+        propagate_ignored(root)
+        return device_nodes
+
+    def get_device_tree_filtered(self):
+        """Get a tree of all handleable devices."""
+        root = self.get_device_tree()['/']
+        # remove ignored devices from the structure:
+        def filter_handled(node):
+            node.children[:] = [
+                descendant
+                for child in node.children
+                for descendant in filter_handled(child)
+            ]
+            if node.ignored:
+                return node.children
+            else:
+                return [dev]
+        filter_handled(root)
+        return root
+
+
+class DevNode:
+
+    def __init__(self, device, root, children, ignored):
+        self.device = device
+        self.root = root
+        self.children = children
+        self.ignored = ignored
+
+    def _sort_key(self):
+        return self.device.device_presentation if self.device else ''
 
 
 # data structs containing the menu hierarchy:
 Device = namedtuple('Device', ['root', 'branches', 'device', 'label', 'methods'])
 Action = namedtuple('Action', ['method', 'device', 'label', 'action'])
-Branch = namedtuple('Branch', ['label', 'groups'])
 
 
 class DeviceActions(object):
@@ -608,9 +827,10 @@ class DeviceActions(object):
         'unmount': _('Unmount {0}'),
         'unlock': _('Unlock {0}'),
         'lock': _('Lock {0}'),
-        'eject': _('Eject {0}'),
-        'detach': _('Unpower {0}'),
+        'eject': _('Eject {1}'),
+        'detach': _('Unpower {1}'),
         'forget_password': _('Clear password for {0}'),
+        'delete': _('Detach {0}'),
     }
 
     def __init__(self, mounter, actions={}):
@@ -625,9 +845,10 @@ class DeviceActions(object):
             'eject': partial(mounter.eject, force=True),
             'detach': partial(mounter.detach, force=True),
             'forget_password': mounter.forget_password,
+            'delete': mounter.delete,
         })
 
-    def detect(self, root_device=''):
+    def detect(self, root_device='/'):
         """
         Detect all currently known devices.
 
@@ -639,17 +860,19 @@ class DeviceActions(object):
         device_nodes = dict(map(self._device_node,
                                 self._mounter.get_all_handleable()))
         # insert child devices as branches into their roots:
-        for object_path, node in device_nodes.items():
+        for node in device_nodes.values():
             device_nodes.get(node.root, root).branches.append(node)
-        if not root_device:
-            return root
+        device_nodes['/'] = root
+        for node in device_nodes.values():
+            node.branches.sort(key=lambda node: node.label)
         return device_nodes[root_device]
 
     def _get_device_methods(self, device):
         """Return an iterable over all available methods the device has."""
         if device.is_filesystem:
             if device.is_mounted:
-                yield 'browse'
+                if self._mounter._browser:
+                    yield 'browse'
                 yield 'unmount'
             else:
                 yield 'mount'
@@ -665,24 +888,22 @@ class DeviceActions(object):
             yield 'eject'
         if device.is_detachable:
             yield 'detach'
+        if device.is_loop and device.loop_support:
+            yield 'delete'
 
     def _device_node(self, device):
         """Create an empty menu node for the specified device."""
         label = device.ui_label
+        dev_label = device.ui_device_label
         # determine available methods
         methods = [Action(method, device,
-                          self._labels[method].format(label),
+                          self._labels[method].format(label, dev_label),
                           partial(self._actions[method], device))
                    for method in self._get_device_methods(device)]
         # find the root device:
-        if device.is_partition:
-            root = device.partition_slave.object_path
-        elif device.is_luks_cleartext:
-            root = device.luks_cleartext_slave.object_path
-        else:
-            root = None
+        root = device.parent_object_path
         # in this first step leave branches empty
-        return device.object_path, Device(root, [], device, label, methods)
+        return device.object_path, Device(root, [], device, dev_label, methods)
 
 
 def prune_empty_node(node, seen):
