@@ -18,9 +18,9 @@ import logging
 
 from gi.repository import GLib
 
-from .common import Emitter, samefile, AttrDictView, decode_ay
-from .compat import fix_str_conversions
-from .dbus import connect_service, MethodsProxy
+from .common import (Emitter, samefile, sameuuid, AttrDictView, decode_ay,
+                     BaseDevice)
+from .dbus import connect_service, MethodsProxy, DBusCallWithFdList, DBusCall
 from .locale import _
 from .async_ import Coroutine, Return
 
@@ -117,8 +117,7 @@ class PropertiesNotAvailable(object):
 # Device wrapper
 # ----------------------------------------
 
-@fix_str_conversions
-class Device(object):
+class Device(BaseDevice):
 
     """
     Proxy class for UDisks2 devices.
@@ -135,24 +134,6 @@ class Device(object):
         self._P = property_hub
         self._M = method_hub
 
-    def __str__(self):
-        """Show as object_path."""
-        return self.object_path
-
-    def __eq__(self, other):
-        """Comparison by object_path."""
-        return self.object_path == str(other)
-
-    def __ne__(self, other):
-        """Comparison by object_path."""
-        return not (self == other)
-
-    def is_file(self, path):
-        """Comparison by mount and device file path."""
-        return (samefile(path, self.device_file) or
-                samefile(path, self.loop_file) or
-                any(samefile(path, mp) for mp in self.mount_paths))
-
     # availability of interfaces
     @property
     def is_drive(self):
@@ -210,7 +191,9 @@ class Device(object):
         This method is used internally to unify the behaviour of top level
         devices in udisks1 and udisks2.
         """
-        return self.drive if self.is_toplevel and not self.is_loop else self
+        # NOTE: always fallback to `self` because udisks2 doesn't report
+        # CryptoBackingDevice nor Drive for logical volumes:
+        return self.is_toplevel and not self.is_loop and self.drive or self
 
     @property
     def is_detachable(self):
@@ -227,6 +210,16 @@ class Device(object):
         """Check if there is media available in the drive."""
         return bool(self._assocdrive._P.Drive.MediaAvailable)
 
+    @property
+    def drive_vendor(self):
+        """Return drive vendor."""
+        return self._assocdrive._P.Drive.Vendor
+
+    @property
+    def drive_model(self):
+        """Return drive model."""
+        return self._assocdrive._P.Drive.Model
+
     # Drive methods
     def eject(self, auth_no_user_interaction=None):
         """Eject media from the device."""
@@ -392,6 +385,13 @@ class Device(object):
         """Return the recommended device symbolic icon name."""
         return self._P.Block.HintSymbolicIconName or 'drive-removable-media'
 
+    @property
+    def symlinks(self):
+        """Known symlinks of the block device."""
+        if not self._P.Block.Symlinks:
+            return []
+        return [decode_ay(path) for path in self._P.Block.Symlinks]
+
     # ----------------------------------------
     # Partition
     # ----------------------------------------
@@ -402,6 +402,11 @@ class Device(object):
         """Get the partition slave (container)."""
         return self._daemon[self._P.Partition.Table]
 
+    @property
+    def partition_uuid(self):
+        """Get the partition UUID."""
+        return self._P.Partition.UUID
+
     # ----------------------------------------
     # Filesystem
     # ----------------------------------------
@@ -473,6 +478,14 @@ class Device(object):
             })
         )
 
+    def unlock_keyfile(self, password, auth_no_user_interaction=None):
+        return self._M.Encrypted.Unlock(
+            '(sa{sv})', '', filter_opt({
+                'keyfile_contents': ('ay', password),
+                'auth.no_user_interaction': ('b', auth_no_user_interaction),
+            })
+        )
+
     def lock(self, auth_no_user_interaction=None):
         """Lock Luks device."""
         return self._M.Encrypted.Lock(
@@ -491,31 +504,47 @@ class Device(object):
         """Get the file backing the loop device."""
         return decode_ay(self._P.Loop.BackingFile)
 
+    @property
+    def setup_by_uid(self):
+        """Get the ID of the user who set up the loop device."""
+        return self._P.Loop.SetupByUID
+
+    @property
+    def autoclear(self):
+        """If True the loop device will be deleted after unmounting."""
+        return self._P.Loop.Autoclear
+
+    def delete(self, auth_no_user_interaction=None):
+        """Delete loop partition."""
+        return self._M.Loop.Delete(
+            '(a{sv})',
+            filter_opt({
+                'auth.no_user_interaction': ('b', auth_no_user_interaction),
+            })
+        )
+
+    def set_autoclear(self, value, auth_no_user_interaction=None):
+        """Set autoclear flag for loop partition."""
+        return self._M.Loop.SetAutoclear(
+            '(ba{sv})',
+            value,
+            filter_opt({
+                'auth.no_user_interaction': ('b', auth_no_user_interaction),
+            })
+        )
+
+    loop_support = True
+
     # ----------------------------------------
     # derived properties
     # ----------------------------------------
 
-    @property
-    def in_use(self):
-        """Check whether this device is in use, i.e. mounted or unlocked."""
-        if self.is_mounted or self.is_unlocked:
-            return True
-        if self.is_partition_table:
-            for device in self._daemon:
-                if device.partition_slave == self and device.in_use:
-                    return True
-        return False
-
     @property
     def parent_object_path(self):
         return (self._P.Partition.Table
                 or self._P.Block.CryptoBackingDevice
                 or '/')
 
-    @property
-    def ui_label(self):
-        return self.id_label or self.id_uuid or self.device_presentation
-
 
 # ----------------------------------------
 # UDisks2 service wrapper
@@ -564,7 +593,7 @@ class Daemon(Emitter):
                 return device
         raise ValueError(_('no device found owning "{0}"', path))
 
-    def __init__(self, proxy):
+    def __init__(self, proxy, version):
 
         """Initialize object and start listening to UDisks2 events."""
 
@@ -580,8 +609,15 @@ class Daemon(Emitter):
                        'job_failed']
         super(Daemon, self).__init__(event_names)
 
-        self._proxy = proxy
         self._log = logging.getLogger(__name__)
+        self._log.debug(_('Daemon version: {0}', version))
+
+        self.version = version
+        self.version_info = tuple(map(int, version.split('.')))
+        self.keyfile_support = self.version_info >= (2,6,4)
+        self._log.debug(_('Keyfile support: {0}', self.keyfile_support))
+
+        self._proxy = proxy
         self._objects = {}
 
         proxy.connect('InterfacesAdded', self._interfaces_added)
@@ -608,11 +644,46 @@ class Daemon(Emitter):
     @classmethod
     @Coroutine.from_generator_function
     def create(cls):
-        proxy = yield connect_service(cls)
-        daemon = cls(proxy)
+        service = (cls.BusName, cls.ObjectPath, cls.Interface)
+        proxy = yield connect_service(*service)
+        version = yield cls.get_version()
+        daemon = cls(proxy, version)
         yield daemon._sync()
         yield Return(daemon)
 
+    @classmethod
+    @Coroutine.from_generator_function
+    def get_version(cls):
+        service = (cls.BusName,
+                   '/org/freedesktop/UDisks2/Manager',
+                   Interface['Properties'])
+        manager = yield connect_service(*service)
+        version = yield DBusCall(manager._proxy, 'Get', '(ss)', (
+            Interface['Manager'], 'Version'))
+        yield Return(version)
+
+    @Coroutine.from_generator_function
+    def loop_setup(self, fd, options):
+        service = (self.BusName,
+                   '/org/freedesktop/UDisks2/Manager',
+                   Interface['Manager'])
+        manager = yield connect_service(*service)
+        object_path = yield DBusCallWithFdList(
+            manager._proxy, 'LoopSetup', '(ha{sv})',
+            (0, filter_opt({
+                'auth.no_user_interaction': ('b', options.get('auth.no_user_interaction')),
+                'offset': ('t', options.get('offset')),
+                'size': ('t', options.get('size')),
+                'read-only': ('b', options.get('read-only')),
+                'no-part-scan': ('b', options.get('no-part-scan')),
+            })),
+            [fd],
+        )
+        yield self._sync()
+        yield Return(self[object_path])
+
+    loop_support = True
+
     # UDisks2 interface
     def paths(self):
         return self._objects.keys()
